using Wombat
using Plots

setplots()

function showEntropyWave()

    nzones = [8,16,32,64,128,256,512,1024]

    L1 = zeros(Float64, size(nzones,1))

    for i = 1:size(nzones,1)

        N = nzones[i]

        fname = "entr_"*itoa(N,1)*"-0000-000"

        rho, sim = ReadSnap(fname, "RHO")
        vx,  sim = ReadSnap(sim, "VX")
        vy,  sim = ReadSnap(sim, "VY")
        vz,  sim = ReadSnap(sim, "VZ")
        Bx,  sim = ReadSnap(sim, "BX")
        By,  sim = ReadSnap(sim, "BY")
        Bz,  sim = ReadSnap(sim, "BZ")
        P,   sim = ReadSnap(sim, "PRS")

        q0 = [rho vx vy vz Bx By Bz P]

        fname = "entr_"*itoa(N,1)*"-0001-000"
    
        rho, sim = ReadSnap(fname, "RHO")
        vx,  sim = ReadSnap(sim, "VX")
        vy,  sim = ReadSnap(sim, "VY")
        vz,  sim = ReadSnap(sim, "VZ")
        Bx,  sim = ReadSnap(sim, "BX")
        By,  sim = ReadSnap(sim, "BY")
        Bz,  sim = ReadSnap(sim, "BZ")
        P,   sim = ReadSnap(sim, "PRS")

        q1 = [rho vx vy vz Bx By Bz P]
    
        L1[i] = L1_Error(q0,q1)
    end

    plot(nzones, L1, show=true, xscale=:log10, yscale=:log10)

    Lideal = 1.0./Float64.(nzones).^5

    Lideal = Lideal / maximum(Lideal) * 1e-8

    plot!(nzones, Lideal)

end

function L1_Error(q0::Array{Float64,2}, q1::Array{Float64,2})
    
    dq = zeros(Float64, 8)

    N = size(q0, 1)

    for i = 1:8
        dq[1] = sum(abs.(q1[:,1] - q0[:,1])) / N
        dq[2] = sum(abs.(q1[:,2] - q0[:,2])) / N
        dq[3] = sum(abs.(q1[:,3] - q0[:,3])) / N
        dq[4] = sum(abs.(q1[:,4] - q0[:,4])) / N
        dq[5] = sum(abs.(q1[:,5] - q0[:,5])) / N
        dq[6] = sum(abs.(q1[:,6] - q0[:,6])) / N
        dq[7] = sum(abs.(q1[:,7] - q0[:,7])) / N
        dq[8] = sum(abs.(q1[:,8] - q0[:,8])) / N
    end

    return sqrt( sum(dq.^2) )

end
