
using Wombat
using LaTeXStrings
using RayCaster
using Printf

import PGFPlots
using PGFTools

homedir = "/Users/jdonnert/Volumes/iCloud/Projects/2018/WENO-Wombat/"
datadir = "/Users/jdonnert/data/WENOWombat/"
imgdir  = homedir*"img/"


function plot_waves_full_MHD_1d(fout)

    nzones = [8,16,32,64,128,256,512,1024]
    names  = ["alfven","bentropy","fast","slow"]

    L1 = zeros(Float64, size(nzones,1), size(names,1))
    
    for j=1:size(names,1)
    
        for i = 1:size(nzones,1)
            
            N = nzones[i]
        
            fname = datadir*"1D_Waves/"*names[j]*"_"*itoa(N,1)*"-0000-000"

            rho, sim = ReadSnap(fname, "RHO")
            vx,  sim = ReadSnap(sim, "VX")
            vy,  sim = ReadSnap(sim, "VY")
            vz,  sim = ReadSnap(sim, "VZ")
            Bx,  sim = ReadSnap(sim, "BX")
            By,  sim = ReadSnap(sim, "BY")
            Bz,  sim = ReadSnap(sim, "BZ")
            P,   sim = ReadSnap(sim, "PRS")
    
            q0 = [rho vx vy vz Bx By Bz P]

            fname = datadir*"1D_Waves/"*names[j]*"_"*itoa(N,1)*"-0001-000"
    
            rho, sim = ReadSnap(fname, "RHO")
            vx,  sim = ReadSnap(sim, "VX")
            vy,  sim = ReadSnap(sim, "VY")
            vz,  sim = ReadSnap(sim, "VZ")
            Bx,  sim = ReadSnap(sim, "BX")
            By,  sim = ReadSnap(sim, "BY")
            Bz,  sim = ReadSnap(sim, "BZ")
            P,   sim = ReadSnap(sim, "PRS")

            q1 = [rho vx vy vz Bx By Bz P]
    
            L1[i, j] = L1_Error(q0,q1)
        end
    end

    ideal = Float64.(nzones).^(-5) ./ maximum(Float64.(nzones).^(-5)) .* maximum(L1) * 5

    l1 = PGFPlots.Plots.Linear(nzones, L1[:,1], style="Set11,solid", markSize=1,mark="o",legendentry=L"Alfv\'{e}n $$")
    l2 = PGFPlots.Plots.Linear(nzones, L1[:,2], style="Set12,solid", markSize=1,mark="x",legendentry="Entropy")
    l3 = PGFPlots.Plots.Linear(nzones, L1[:,3], style="Set13,solid", markSize=1,mark="square",legendentry="Fast")
    l4 = PGFPlots.Plots.Linear(nzones, L1[:,4], style="Set14,solid", markSize=1,mark="diamond",legendentry="Slow")

    li = PGFPlots.Plots.Linear(nzones, ideal, style="black,dotted", markSize=0,legendentry=L"N^{-5}")

    ax = PGFPlots.Axis([l1,l2,l3,l4,li], xmin=6,xmax=1128, ymin=1e-14, ymax=3e-7, ymode="log", xmode="log", 
                                         xlabel="Number of Zones", ylabel="L1 Error")

    PGFPlots.save(fout, ax)
end

function  show_OT_comparison(fout)

    i = 10

    fname = datadir*"2D_OT/WENO_comp-"*itoa(i,4)*"-000"
    note = PGFPlots.Node("t = "*gtoa(i*0.05),-0.4,-0.4, style="right, white")
    note2 = PGFPlots.Node("WENO5",-0.4,0.4, style="right, white")

    rho, sim = ReadSnap(fname, "RHO")
        
    pgf_img = PGFImage(rho, (-0.5,0.5), (-0.5,0.5), cmap="L8")
    ax = PGFPlots.Axis([pgf_img,note,note2], axisEqualImage=true,xlabel="x", ylabel="y",xmin=-0.5,xmax=0.5,
                       style="grid style={none}")

    fname = datadir*"2D_OT/TVD-"*itoa(i,4)*"-000"
    note3 = PGFPlots.Node("TVD+CTU",-0.4,0.4, style="right, white")

    rho, sim = ReadSnap(fname, "RHO")
        
    pgf_img = PGFImage(rho, (-0.5,0.5), (-0.5,0.5), cmap="L8")
    ax2 = PGFPlots.Axis([pgf_img, note3], axisEqualImage=true,xlabel="x", ylabel="",xmin=-0.5,xmax=0.5,
                       style="yticklabels={},grid style={none},colorbar style={ylabel=Density,width=0.3cm,xshift=-0.15cm}")

	g = PGFPlots.GroupPlot(2, 1, groupStyle = "vertical sep = 0.0cm, horizontal sep = 0.0cm")
    
    push!(g, ax)
    push!(g, ax2)

    PGFPlots.save(fout, g)
end

function plot_OT_comp(i, fout)

    fname = datadir*"2D_OT/WENO_comp-"*itoa(i,4)*"-000"
    P, sim  = ReadSnap(fname, "PRS")

    pos = MakeZonePos(sim)
 
    pos .-= 1/size(P,2) / 2
    good = findall( pos[:,:,2] .+ 0.1875 .== amin(abs.(pos[:,:,2].+0.1875)))

    x = range(-0.5,stop=0.5,length=size(P,1))
    y = P[good]

    l1 = PGFPlots.Plots.Linear(x, y, style="black", markSize=1,mark="square",
                                     legendentry=L"WENO5, $128^2$ zones",onlyMarks=true)
    
    good = findall( pos[:,:,2].+0.073 .== amin(abs.(pos[:,:,2].+0.073)))
    y = P[good]
        
    note = PGFPlots.Node(L"$t = 0.5, y = -0.073$", -0.475, 0.0275, style="right")
    l3 = PGFPlots.Plots.Linear(x, y, style="black", markSize=1, mark="square",
                                     legendentry=L"WENO5, $128^2$ zones",onlyMarks=true)

    fname = datadir*"2D_OT/TVD-"*itoa(i,4)*"-000"
    P, sim  = ReadSnap(fname, "PRS")

    pos = MakeZonePos(sim)
 
    pos .-= 1/size(P,2) / 2
    good = findall( pos[:,:,2].+0.1875 .== amin(abs.(pos[:,:,2].+0.1875)))

    x = range(-0.5,stop=0.5,length=size(P,1))
    y = P[good]

    note1 = PGFPlots.Node(L"$t = 0.5, y = -0.1875$", -0.475, 0.025, style="right")
    l2 = PGFPlots.Plots.Linear(x, y, style="Set11,solid", markSize=0,legendentry=L"TVD+CTU, $256^2$ zones")

    good = findall( Float32.(pos[:,:,2] .+ 0.073) .== Float32(amin( abs.(pos[:,:,2] .+ 0.073 ))) )
    y = P[good]
        
    l4 = PGFPlots.Plots.Linear(x, y, style="Set11,solid", markSize=0,
                                     legendentry=L"TVD+CTU, $256^2$ zones")

    ax1 = PGFPlots.Axis([l1,l2, note1], xmin=-0.5,xmax=0.5, ymin=0.0, ymax=0.5,
                                         xlabel="x", ylabel="Pressure ")

    ax2 = PGFPlots.Axis([l3,l4, note], xmin=-0.5,xmax=0.5, ymin=0.0, ymax=0.5,
                        xlabel="x", ylabel="",style="yticklabels={}")


	g = PGFPlots.GroupPlot(2, 1, groupStyle = "vertical sep = 0.5cm, horizontal sep = 0.7cm")
    
    push!(g, ax1)
    push!(g, ax2)

    PGFPlots.save(fout, g)
    
end

function plot_field_loop_energy_2d(fout)

    N = 40

    B2_arr = zeros(Float64, N, 6)

    for snap = 1:N

        fname = datadir*"2D_FieldLoop/WENO_32-"*itoa(snap,4)*"-000"

        Bx, sim = ReadSnap(fname, "BX")
        By, sim = ReadSnap(sim, "BY")
        Bz, sim = ReadSnap(sim, "BZ")

        B2_arr[snap, 1] = sum(Bx.^2 + By.^2 + Bz.^2)
    
        fname = datadir*"2D_FieldLoop/WENO_64-"*itoa(snap,4)*"-000"

        Bx, sim = ReadSnap(fname, "BX")
        By, sim = ReadSnap(sim, "BY")
        Bz, sim = ReadSnap(sim, "BZ")

        B2_arr[snap, 2] = sum(Bx.^2 + By.^2 + Bz.^2)

        fname = datadir*"2D_FieldLoop/WENO_128-"*itoa(snap,4)*"-000"

        Bx, sim = ReadSnap(fname, "BX")
        By, sim = ReadSnap(sim, "BY")
        Bz, sim = ReadSnap(sim, "BZ")

        B2_arr[snap, 3] = sum(Bx.^2 + By.^2 + Bz.^2)

        fname = datadir*"2D_FieldLoop/TVD_128-"*itoa(snap,4)*"-000"

        Bx, sim = ReadSnap(fname, "BX")
        By, sim = ReadSnap(sim, "BY")
        Bz, sim = ReadSnap(sim, "BZ")

        B2_arr[snap, 4] = sum(Bx.^2 + By.^2 + Bz.^2)
    
        fname = datadir*"2D_FieldLoop/TVD_64-"*itoa(snap,4)*"-000"

        Bx, sim = ReadSnap(fname, "BX")
        By, sim = ReadSnap(sim, "BY")
        Bz, sim = ReadSnap(sim, "BZ")

        B2_arr[snap, 5] = sum(Bx.^2 + By.^2 + Bz.^2)

        fname = datadir*"2D_FieldLoop/TVD_32-"*itoa(snap,4)*"-000"

        Bx, sim = ReadSnap(fname, "BX")
        By, sim = ReadSnap(sim, "BY")
        Bz, sim = ReadSnap(sim, "BZ")

        B2_arr[snap, 6] = sum(Bx.^2 + By.^2 + Bz.^2)

    end

    t = linspace(0,N*0.05, N)

    B2_arr[:,1] /= B2_arr[1,1]
    B2_arr[:,2] /= B2_arr[1,2]
    B2_arr[:,3] /= B2_arr[1,3]

    B2_arr[:,4] /= B2_arr[1,4]
    B2_arr[:,5] /= B2_arr[1,5]
    B2_arr[:,6] /= B2_arr[1,6]

    l1 = PGFPlots.Plots.Linear(t, B2_arr[:,1], style="Set11,solid", markSize=0,
                               legendentry="WENO5, 32 zones")
    l2 = PGFPlots.Plots.Linear(t, B2_arr[:,2], style="Set12,solid", markSize=0,
                               legendentry="WENO5, 64 zones")
    l3 = PGFPlots.Plots.Linear(t, B2_arr[:,3], style="Set13,solid", markSize=0,
                               legendentry="WENO5, 128 zones")

    l4 = PGFPlots.Plots.Linear(t, B2_arr[:,4], style="Set13,dashed", markSize=0,
                               legendentry="TVD+CTU, 128 zones")
    l5 = PGFPlots.Plots.Linear(t, B2_arr[:,5], style="Set12,dashed", markSize=0,
                               legendentry="TVD+CTU, 64 zones")
    l6 = PGFPlots.Plots.Linear(t, B2_arr[:,6], style="Set11,dashed", markSize=0,
                               legendentry="TVD+CTU, 32 zones")

    l0 = PGFPlots.Plots.Linear([0,2], [0.98,0.98], style="gray,dotted", markSize=0,
                              legendentry="98 \\%")

    ax = PGFPlots.Axis([l1,l2,l3,l4,l5,l6, l0], xmin=0,xmax=maximum(t), ymin=0.5, ymax=1.0, xlabel="Time",
                       ylabel=L"\mathrm{B}^2", legendStyle="at={(0.5,0.5)},anchor=west")


    PGFPlots.save(fout, ax)

end

function show_Kelvin_Helmholtz_comp(snap, fout)

    t = snap * 0.1
    
    range = [0.95,2.05]

    note1 = PGFPlots.Node("t = "*gtoa(t),0.2,-0.425, style="right,white,font={\\footnotesize}")
    note  = PGFPlots.Node(L"WENO5, $64 \times 128$ zones",-0.425,0.425, style="right,white,font={\\footnotesize}")

    fname = datadir*"2D_KH/WENO_64-"*itoa(snap,4)*"-000"
    rho, sim = ReadSnap(fname, "RHO") 

    rho = [ rho;rho ]

    pgf_img = PGFImage(rho,(-0.5,0.5), (-0.5,0.5), cmap="D1A", zlims=range)
    
    ax1 = PGFPlots.Axis([pgf_img, note1,note]; xmin=-0.5,xmax=0.5, ymin=-0.5, ymax=0.5,
                                  axisEqualImage=true,colorbar=false,
      style="grid style={none},colorbar style={ylabel=Density,width=0.3cm,xshift=-0.2cm}")

    # TVD

    fname = datadir*"2D_KH/TVD_128-"*itoa(snap,4)*"-000"
    rho, sim = ReadSnap(fname, "RHO") 

    rho = [ rho;rho ]

    pgf_img = PGFImage(rho,(-0.5,0.5), (-0.5,0.5), cmap="D1A", zlims=range)

    note  = PGFPlots.Node(L"TVD+CTU,  $128 \times 256$ zones",-0.425,0.425, style="right,white,font={\\footnotesize}")

    ax3 = PGFPlots.Axis([pgf_img,note]; xmin=-0.5,xmax=0.5, ymin=-0.5, ymax=0.5, xlabel="x",
                        axisEqualImage=true,
                        style="grid style={none},colorbar style={ylabel=Density,width=0.3cm,xshift=-0.2cm}, yticklabels={none}")
    

	g = PGFPlots.GroupPlot(2, 1, groupStyle = "horizontal sep = 0.1cm, vertical sep = 0.1cm")
    
    push!(g, ax1)
    push!(g, ax3)

    PGFPlots.save(fout, g)
end


function plot_HLRS_weak_scaling(fout)

    nnodes   = [1,2,4,8,16,32,64,128,256,512,1024,2048,4096]
    nthreads = 24
    nstep    = 100
    
    N = length(nnodes)
    walltime = zeros(Float64, N)
    nzones   = zeros(Float64, N)
    stepmin  = zeros(Float64, N)
    stepmax  = zeros(Float64, N)

    for i = 1:N

        tab = readdlm("/Users/jdonnert/data/WENOWombat/weak/log/blob_"*itoa(nnodes[i],1)*".log")

        good = find(tab[:,1] .== "Step")

        str = minimum(tab[good,11])
        stepmin[i]  = parse(Float64, str[1:6])

        str = maximum(tab[good,11])
        stepmax[i]  = parse(Float64, str[1:6])
        
        walltime[i] = maximum(tab[good,14])
    
        good = find(tab[:,1] .== "Total")
        nzones[i] = tab[good[1], 5]

        println(nnodes[i], " ", walltime[i])
    end

    stepmean = walltime/nstep

    errBars = PGFPlots.ErrorBars(yplus=stepmax-stepmean, yminus=stepmean-stepmin)

    l1 = PGFPlots.Plots.Linear(nnodes, stepmean, style="Set11,solid", markSize=1, 
                               legendentry="WENO-Wombat",
                              errorBars=errBars)
    # TVD

    walltime = zeros(Float64, N)
    nzones   = zeros(Float64, N)
    stepmin  = zeros(Float64, N)
    stepmax  = zeros(Float64, N)

    for i = 1:N-1

        tab = readdlm("/Users/jdonnert/data/WENOWombat/weak_tvd/log/blob_"*itoa(nnodes[i],1)*".log")

        good = find(tab[:,1] .== "Step")

        str = minimum(tab[good,11])
        stepmin[i]  = parse(Float64, str[1:6])

        str = maximum(tab[good,11])
        stepmax[i]  = parse(Float64, str[1:6])
        
        walltime[i] = maximum(tab[good,14])
    
        good = find(tab[:,1] .== "Total")
        nzones[i] = tab[good[1], 5]

        println(nnodes[i], " ", walltime[i])
    end

    stepmean = walltime/nstep

    errBars = PGFPlots.ErrorBars(yplus=stepmax-stepmean, yminus=stepmean-stepmin)

    l2 = PGFPlots.Plots.Linear(nnodes, stepmean, style="Set12,solid", markSize=1, 
                               legendentry="TVD+CTU-Wombat",
                              errorBars=errBars)

    lii = PGFPlots.Plots.Linear([1,4096], [2.0,2.0], style="black,dashed", markSize=0, legendentry="ideal, 2.0 sec")
    li = PGFPlots.Plots.Linear([1,4096], [2.45,2.45], style="black,dotted", markSize=0, legendentry="ideal, 2.45 sec")

    l0 = PGFPlots.Plots.Linear([0,0], [0,0], style="black,solid, axis x line={bottom}", markSize=0)
    
    ax = PGFPlots.Axis([l1,l2,li,lii]; xmin=0.8,xmax=5000, ymin=1.8, ymax=2.8, xlabel="Number of Nodes", ylabel="Time per update [sec]", 
                              legendPos="north east", xmode="log", style="axis x line*={bottom},legend columns={2}")
    ax2 = PGFPlots.Axis([l0]; xmin=0.8*24, xmax = 5000*24, ymin=1.8, ymax=2.8, xlabel="Number of Cores",
                        style=", axis y line={none}, axis x line*={top},grid=none", xmode="log")


    PGFPlots.save(fout, [ax, ax2])
    
end

function make_PDFs()
    
    # Images

    PGFPlots.resetPGFPlotsPreamble()

    show_OT_comparison(imgdir*"OrszagTang_comp.pdf")
    show_Kelvin_Helmholtz_comp(20, imgdir*"Kelvin_Helmholtz_Density_comp.pdf")

    # Plots

    InitPGF()

    plot_waves_full_MHD_1d(imgdir*"/MHDWaves_full_1d.pdf")
    plot_OT_comp(10, imgdir*"OrszagTang_Slice_Comp.pdf")
    plot_HLRS_weak_scaling(imgdir*"/WeakScaling_HazelHen.pdf")

end
